// geometry.cpp
// DEFINOVANIE ZAKLADNYCH GEOMETRICKYCH OBJEKTOV

#include "geometry.h"
#include <math.h>
#include <memory.h>


// id cube VBO
GLuint g_uiVBOCubeCoords = 0, g_uiVBOCubeTexCoords = 0, g_uiVBOCubeNormals = 0, g_uiVBOCubeTangents = 0, g_uiVBOCubeIndices = 0;

// id sphere VBO
GLuint g_uiVBOSphereCoords = 0, g_uiVBOSphereTexCoords = 0, g_uiVBOSphereNormals = 0, g_uiVBOSphereTangents = 0, g_uiVBOSphereIndices = 0;
GLuint g_uiVBOSphereNumIndices;

// id torus VBO
GLuint g_uiVBOTorusCoords = 0, g_uiVBOTorusTexCoords = 0, g_uiVBOTorusNormals = 0, g_uiVBOTorusTangents = 0, g_uiVBOTorusIndices = 0;
GLuint g_uiVBOTorusNumIndices;

#define PI 3.14159f
#define NULL 0

GLfloat cube_vertices[] = {1,1,1,   -1,1,1,   -1,-1,1,  1,-1,1,  
                          1,1,1,    1,-1,1,   1,-1,-1,  1,1,-1,
                          1,1,1,    1,1,-1,   -1,1,-1,  -1,1,1,
                          -1,1,1,   -1,1,-1,  -1,-1,-1, -1,-1,1,
                          -1,-1,-1, 1,-1,-1,  1,-1,1,   -1,-1,1,
                          1,-1,-1,  -1,-1,-1, -1,1,-1,  1,1,-1 };
GLfloat cube_normals[] = {0,0,1,    0,0,1,   0,0,1,   0,0,1,  
                          1,0,0,    1,0,0,   1,0,0,   1,0,0,
                          0,1,0,    0,1,0,   0,1,0,   0,1,0,
                          -1,0,0,   -1,0,0,  -1,0,0,  -1,0,0,
                          0,-1,0,   0,-1,0,  0,-1,0,  0,-1,0,
                          0,0,-1,   0,0,-1,  0,0,-1,  0,0,-1};
GLfloat cube_tangents[] = {-1,0,0,1,    -1,0,0,1,   -1,0,0,1,  -1,0,0,1,
                           0,-1,0,1,    0,-1,0,1,   0,-1,0,1,  0,-1,0,1,
                           0,0,-1,1,    0,0,-1,1,   0,0,-1,1,  0,0,-1,1,
                           0,0,-1,1,    0,0,-1,1,   0,0,-1,1,  0,0,-1,1,
                           1,0,0,1,     1,0,0,1,    1,0,0,1,   1,0,0,1,
                           -1,0,0,1,    -1,0,0,1,   -1,0,0,1,  -1,0,0,1};
GLfloat cube_texcoords[] = {0,0,   1,0,  1,1,   0,1,  
                           0,0,   1,0,  1,1,   0,1,
                           0,0,   1,0,  1,1,   0,1,
                           0,0,   1,0,  1,1,   0,1,
                           0,0,   1,0,  1,1,   0,1,
                           0,0,   1,0,  1,1,   0,1};
GLuint cube_indices[] = {0,1,2,3,  4,5,6,7,  8,9,10,11,   12,13,14,15,   16,17,18,19,  20,21,22,23};


// priprava VBO pre renderovanie kocky
void initCubeVBO()
{
    glGenBuffers(1, &g_uiVBOCubeCoords);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOCubeCoords);
    glBufferData(GL_ARRAY_BUFFER, 24*3*sizeof(float), cube_vertices, GL_STATIC_DRAW);

    glGenBuffers(1, &g_uiVBOCubeNormals);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOCubeNormals);
    glBufferData(GL_ARRAY_BUFFER, 24*3*sizeof(float), cube_normals, GL_STATIC_DRAW);

    glGenBuffers(1, &g_uiVBOCubeTexCoords);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOCubeTexCoords);
    glBufferData(GL_ARRAY_BUFFER, 24*2*sizeof(float), cube_texcoords, GL_STATIC_DRAW);

    glGenBuffers(1, &g_uiVBOCubeTangents);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOCubeTangents);
    glBufferData(GL_ARRAY_BUFFER, 24*4*sizeof(float), cube_tangents, GL_STATIC_DRAW);
    
    glGenBuffers(1, &g_uiVBOCubeIndices);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, g_uiVBOCubeIndices);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 6*4*sizeof(GLuint), cube_indices, GL_STATIC_DRAW);
}

// priprava VBO pre renderovanie sfery
void initSphereVBO(int xSlices, int ySlices)
{
    float* vertices = new float[3 * (xSlices + 1) * (ySlices + 1)];
    float* normals = new float[3 * (xSlices + 1) * (ySlices + 1)];
    float* texcoords = new float[2 * (xSlices + 1) * (ySlices + 1)];

    for (int i = 0; i <= xSlices; i++)
    {
        float u = i / float(xSlices);
        for (int j = 0; j <= ySlices; j++)
        {
            float v = j / float(ySlices);

            int index = i + j * (xSlices + 1);

            vertices[3 * index + 0] = cos(2 * u * PI) * cos((v - 0.5f) * PI);
            vertices[3 * index + 1] = sin(2 * u * PI) * cos((v - 0.5f) * PI);
            vertices[3 * index + 2] = sin((v - 0.5f) * PI);

            normals[3 * index + 0] = vertices[3 * index + 0];
            normals[3 * index + 1] = vertices[3 * index + 1];
            normals[3 * index + 2] = vertices[3 * index + 2];

            texcoords[2 * index + 0] = u;
            texcoords[2 * index + 1] = 1.0f - v;
        }
    }

    GLuint* indices = new GLuint[4 * (xSlices + 0) * (ySlices + 0)];
    int index = 0;
    for (int i = 0; i < xSlices; i++)
    {
        for (int j = 0; j < ySlices; j++)
        {
            indices[index] = i + j * (xSlices + 1);
            index++;
            indices[index] = (i + 1) + j * (xSlices + 1);
            index++;
            indices[index] = (i + 1) + (j + 1) * (xSlices + 1);
            index++;
            indices[index] = i + (j + 1) * (xSlices + 1);
            index++;
        }
    }

    glGenBuffers(1, &g_uiVBOSphereCoords);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOSphereCoords);
    glBufferData(GL_ARRAY_BUFFER, 3 * (xSlices + 1) * (ySlices + 1) * sizeof(float), vertices, GL_STATIC_DRAW);

    glGenBuffers(1, &g_uiVBOSphereNormals);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOSphereNormals);
    glBufferData(GL_ARRAY_BUFFER, 3 * (xSlices + 1) * (ySlices + 1) * sizeof(float), normals, GL_STATIC_DRAW);

    glGenBuffers(1, &g_uiVBOSphereTexCoords);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOSphereTexCoords);
    glBufferData(GL_ARRAY_BUFFER, 2 * (xSlices + 1) * (ySlices + 1) * sizeof(float), texcoords, GL_STATIC_DRAW);

    glGenBuffers(1, &g_uiVBOSphereIndices);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, g_uiVBOSphereIndices);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 4 * (xSlices + 0) * (ySlices + 0) * sizeof(GLuint), indices, GL_STATIC_DRAW);

    g_uiVBOSphereNumIndices = 4 * (xSlices + 0) * (ySlices + 0);

    // vypocitaj tangenty
    float* tangents = new float[4 * (xSlices + 1) * (ySlices + 1)];
    CalculateTangentArray((xSlices + 1) * (ySlices + 1), vertices, normals, texcoords, GL_QUADS, 
                          g_uiVBOSphereNumIndices, indices, tangents);
    glGenBuffers(1, &g_uiVBOSphereTangents);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOSphereTangents);
    glBufferData(GL_ARRAY_BUFFER, 4 * (xSlices + 1) * (ySlices + 1) * sizeof(float), tangents, GL_STATIC_DRAW);

    delete [] vertices;
    delete [] normals;
    delete [] texcoords;
    delete [] indices;
    delete [] tangents;
}

// priprava VBO pre renderovanie torusu
void initTorusVBO(float innerRadius, float outerRadius, int xSlices, int ySlices)
{
    float* vertices = new float[3 * (xSlices + 1) * (ySlices + 1)];
    float* normals = new float[3 * (xSlices + 1) * (ySlices + 1)];
    float* texcoords = new float[2 * (xSlices + 1) * (ySlices + 1)];

    for (int i = 0; i <= xSlices; i++)
    {
        float u = i / float(xSlices);
        for (int j = 0; j <= ySlices; j++)
        {
            float v = j / float(ySlices);

            int index = i + j * (xSlices + 1);

            vertices[3 * index + 0] = cos(2 * u * PI) * (outerRadius + innerRadius * cos(2 * v * PI));
            vertices[3 * index + 1] = sin(2 * u * PI) * (outerRadius + innerRadius * cos(2 * v * PI));
            vertices[3 * index + 2] = innerRadius * sin(2 * v * PI);

            float tx = -sin(2 * u * PI);
            float ty = cos(2 * u * PI);
            float tz = 0;
            float sx = cos(2 * u * PI)*(-sin(2 * v * PI));
            float sy = sin(2 * u * PI)*(-sin(2 * v * PI));
            float sz = cos(2 * v * PI);
            normals[3 * index + 0] = ty*sz - tz*sy;
            normals[3 * index + 1] = tz*sx - tx*sz;
            normals[3 * index + 2] = tx*sy - ty*sx;

            texcoords[2 * index + 0] = u;
            texcoords[2 * index + 1] = 1.0f - v;
        }
    }

    GLuint* indices = new GLuint[4 * (xSlices + 0) * (ySlices + 0)];
    int index = 0;
    for (int i = 0; i < xSlices; i++)
    {
        for (int j = 0; j < ySlices; j++)
        {
            indices[index] = i + j * (xSlices + 1);
            index++;
            indices[index] = (i + 1) + j * (xSlices + 1);
            index++;
            indices[index] = (i + 1) + (j + 1) * (xSlices + 1);
            index++;
            indices[index] = i + (j + 1) * (xSlices + 1);
            index++;
        }
    }

    glGenBuffers(1, &g_uiVBOTorusCoords);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOTorusCoords);
    glBufferData(GL_ARRAY_BUFFER, 3 * (xSlices + 1) * (ySlices + 1) * sizeof(float), vertices, GL_STATIC_DRAW);

    glGenBuffers(1, &g_uiVBOTorusNormals);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOTorusNormals);
    glBufferData(GL_ARRAY_BUFFER, 3 * (xSlices + 1) * (ySlices + 1) * sizeof(float), normals, GL_STATIC_DRAW);

    glGenBuffers(1, &g_uiVBOTorusTexCoords);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOTorusTexCoords);
    glBufferData(GL_ARRAY_BUFFER, 2 * (xSlices + 1) * (ySlices + 1) * sizeof(float), texcoords, GL_STATIC_DRAW);

    glGenBuffers(1, &g_uiVBOTorusIndices);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, g_uiVBOTorusIndices);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 4 * (xSlices + 0) * (ySlices + 0) * sizeof(GLuint), indices, GL_STATIC_DRAW);

    g_uiVBOTorusNumIndices = 4 * (xSlices + 0) * (ySlices + 0);

    // vypocitaj tangenty
    float* tangents = new float[4 * (xSlices + 1) * (ySlices + 1)];
    CalculateTangentArray((xSlices + 1) * (ySlices + 1), vertices, normals, texcoords, GL_QUADS, 
                          g_uiVBOTorusNumIndices, indices, tangents);
    glGenBuffers(1, &g_uiVBOTorusTangents);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOTorusTangents);
    glBufferData(GL_ARRAY_BUFFER, 4 * (xSlices + 1) * (ySlices + 1) * sizeof(float), tangents, GL_STATIC_DRAW);

    delete [] vertices;
    delete [] normals;
    delete [] texcoords;
    delete [] indices;
    delete [] tangents;
}

// vykresli jednotkovu kocku s pomocou pripravenych VBO
void renderCubeVBO(GLuint shader_program)
{
    if (g_uiVBOCubeCoords == 0)
        initCubeVBO();

    glEnableClientState(GL_VERTEX_ARRAY);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOCubeCoords);
    glVertexPointer(3, GL_FLOAT, 0, NULL);

    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOCubeTexCoords);
    glTexCoordPointer(2, GL_FLOAT, 0, NULL);

    glEnableClientState(GL_NORMAL_ARRAY);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOCubeNormals);
    glNormalPointer(GL_FLOAT, 0, NULL);

    GLint tangent_attribute_index = glGetAttribLocation(shader_program, "tangent");
    if (g_uiVBOCubeTangents > 0 && tangent_attribute_index > -1)
    {
        glEnableVertexAttribArray(tangent_attribute_index);
        glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOCubeTangents);
        glVertexAttribPointer(tangent_attribute_index, 4, GL_FLOAT, GL_FALSE, 0, 0);
    }

    // draw a cube using buffer with indices
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, g_uiVBOCubeIndices);
    glDrawElements(GL_QUADS, 24, GL_UNSIGNED_INT, NULL);

    // deactivate vertex arrays after drawing
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableVertexAttribArray(tangent_attribute_index);

    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

// vykresli jednotkovu sferu s pomocou pripravenych VBO
void renderSphereVBO(GLuint shader_program)
{
    if (g_uiVBOSphereCoords == 0)
        initSphereVBO(16, 16);

    glEnableClientState(GL_VERTEX_ARRAY);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOSphereCoords);
    glVertexPointer(3, GL_FLOAT, 0, NULL);

    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOSphereTexCoords);
    glTexCoordPointer(2, GL_FLOAT, 0, NULL);

    glEnableClientState(GL_NORMAL_ARRAY);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOSphereNormals);
    glNormalPointer(GL_FLOAT, 0, NULL);

    GLint tangent_attribute_index = glGetAttribLocation(shader_program, "tangent");
    if (g_uiVBOSphereTangents > 0 && tangent_attribute_index > -1)
    {
        glEnableVertexAttribArray(tangent_attribute_index);
        glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOSphereTangents);
        glVertexAttribPointer(tangent_attribute_index, 4, GL_FLOAT, GL_FALSE, 0, 0);
    }

    // draw a cube using buffer with indices
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, g_uiVBOSphereIndices);
    glDrawElements(GL_QUADS, g_uiVBOSphereNumIndices, GL_UNSIGNED_INT, NULL);

    // deactivate vertex arrays after drawing
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableVertexAttribArray(tangent_attribute_index);

    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

// vykresli torus s pomocou pripravenych VBO
void renderTorusVBO(GLuint shader_program)
{
    if (g_uiVBOTorusCoords == 0)
        initTorusVBO(2, 0.2f, 16, 16);

    glEnableClientState(GL_VERTEX_ARRAY);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOTorusCoords);
    glVertexPointer(3, GL_FLOAT, 0, NULL);

    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOTorusTexCoords);
    glTexCoordPointer(2, GL_FLOAT, 0, NULL);

    glEnableClientState(GL_NORMAL_ARRAY);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOTorusNormals);
    glNormalPointer(GL_FLOAT, 0, NULL);

    GLint tangent_attribute_index = glGetAttribLocation(shader_program, "tangent");
    if (g_uiVBOTorusTangents > 0 && tangent_attribute_index > -1)
    {
        glEnableVertexAttribArray(tangent_attribute_index);
        glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOTorusTangents);
        glVertexAttribPointer(tangent_attribute_index, 4, GL_FLOAT, GL_FALSE, 0, 0);
    }

    // draw a cube using buffer with indices
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, g_uiVBOTorusIndices);
    glDrawElements(GL_QUADS, g_uiVBOTorusNumIndices, GL_UNSIGNED_INT, NULL);

    // deactivate vertex arrays after drawing
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableVertexAttribArray(tangent_attribute_index);

    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

// vycitime alokovane objekty
void DeleteAllVBO()
{
    if (g_uiVBOSphereCoords > 0)
        glDeleteBuffers(1, &g_uiVBOSphereCoords);
    if (g_uiVBOSphereTexCoords > 0)
        glDeleteBuffers(1, &g_uiVBOSphereTexCoords);
    if (g_uiVBOSphereNormals > 0)
        glDeleteBuffers(1, &g_uiVBOSphereNormals);
    if (g_uiVBOSphereIndices > 0)
        glDeleteBuffers(1, &g_uiVBOSphereIndices);

    if (g_uiVBOTorusCoords > 0)
        glDeleteBuffers(1, &g_uiVBOTorusCoords);
    if (g_uiVBOTorusTexCoords > 0)
        glDeleteBuffers(1, &g_uiVBOTorusTexCoords);
    if (g_uiVBOTorusNormals > 0)
        glDeleteBuffers(1, &g_uiVBOTorusNormals);
    if (g_uiVBOTorusIndices > 0)
        glDeleteBuffers(1, &g_uiVBOTorusIndices);
}

void CalculateTangentArray(GLuint num_vertices, float* vertices_3D, float* normals_3D,
        float* texcoords_2D, GLenum polygon_mode, int num_indices, GLuint* indices, float* final_tangents_4D)
{
    // kod zalozeny na clanku http://www.terathon.com/code/tangent.html
    // pripravime pomocne polia, ktore budu vo vrcholoch akumulovat tangentu a bitangentu
    float* tangents = new float[3 * num_vertices * 2];
    float* bitangents = tangents + 3 * num_vertices;
    memset(tangents, 0, 3 * sizeof(float) * num_vertices * 2);
    
    int num_polygon_indices = 3;
    if (polygon_mode == GL_QUADS)
        num_polygon_indices = 4;
    for (int polygon = 0; polygon < num_indices / num_polygon_indices; polygon++)
    {
        // ked mame stvoruholnik, musime pocitat 2 trojuholniky, preto tento cyklus
        for (int a = 0; a <= (num_polygon_indices - 3); a++)
        {
            // ziskaj indexy vrcholov aktualneho trojuholnika
            int i1 = indices[num_polygon_indices * polygon + 0];
            int i2 = indices[num_polygon_indices * polygon + a + 1];
            int i3 = indices[num_polygon_indices * polygon + a + 2];
            
            // ziskaj suradnice troch vrcholov trojuholnika
            float v1x = vertices_3D[3 * i1];
            float v1y = vertices_3D[3 * i1 + 1];
            float v1z = vertices_3D[3 * i1 + 2];
            float v2x = vertices_3D[3 * i2];
            float v2y = vertices_3D[3 * i2 + 1];
            float v2z = vertices_3D[3 * i2 + 2];
            float v3x = vertices_3D[3 * i3];
            float v3y = vertices_3D[3 * i3 + 1];
            float v3z = vertices_3D[3 * i3 + 2];
            
            // ziskaj texturne suradnice troch vrcholov trojuholnika
            float w1x = texcoords_2D[2 * i1];
            float w1y = texcoords_2D[2 * i1 + 1];
            float w2x = texcoords_2D[2 * i2];
            float w2y = texcoords_2D[2 * i2 + 1];
            float w3x = texcoords_2D[2 * i3];
            float w3y = texcoords_2D[2 * i3 + 1];
            
            // vypocitaj tangentu a bitangentu
            float x1 = v2x - v1x;
            float x2 = v3x - v1x;
            float y1 = v2y - v1y;
            float y2 = v3y - v1y;
            float z1 = v2z - v1z;
            float z2 = v3z - v1z;
            
            float s1 = w2x - w1x;
            float s2 = w3x - w1x;
            float t1 = w2y - w1y;
            float t2 = w3y - w1y;
            
            float r = 1.0F / (s1 * t2 - s2 * t1);
            float tangent_x = (t2 * x1 - t1 * x2) * r;
            float tangent_y = (t2 * y1 - t1 * y2) * r;
            float tangent_z = (t2 * z1 - t1 * z2) * r;
            float bitangent_x = (s1 * x2 - s2 * x1) * r;
            float bitangent_y = (s1 * y2 - s2 * y1) * r;
            float bitangent_z = (s1 * z2 - s2 * z1) * r;
            
            // naakumuluj vypocitanu tangentu
            tangents[3 * i1 + 0] += tangent_x;
            tangents[3 * i1 + 1] += tangent_y;
            tangents[3 * i1 + 2] += tangent_z;
            tangents[3 * i2 + 0] += tangent_x;
            tangents[3 * i2 + 1] += tangent_y;
            tangents[3 * i2 + 2] += tangent_z;
            tangents[3 * i3 + 0] += tangent_x;
            tangents[3 * i3 + 1] += tangent_y;
            tangents[3 * i3 + 2] += tangent_z;
            
            bitangents[3 * i1 + 0] += bitangent_x;
            bitangents[3 * i1 + 1] += bitangent_y;
            bitangents[3 * i1 + 2] += bitangent_z;
            bitangents[3 * i2 + 0] += bitangent_x;
            bitangents[3 * i2 + 1] += bitangent_y;
            bitangents[3 * i2 + 2] += bitangent_z;
            bitangents[3 * i3 + 0] += bitangent_x;
            bitangents[3 * i3 + 1] += bitangent_y;
            bitangents[3 * i3 + 2] += bitangent_z;
        }
    }
    
    for (int a = 0; a < num_vertices; a++)
    {        
        // pouzijeme Gramov-Schmidtov ortogonalizacny proces, aby sme spravili tangentu kolmu na normalu
        // tu predpokladame, ze tangenta je "skoro" kolma na normalu
        float dot = tangents[3*a+0]*normals_3D[3*a+0] + tangents[3*a+1]*normals_3D[3*a+1] + tangents[3*a+2]*normals_3D[3*a+2];
        final_tangents_4D[4*a+0] = tangents[3*a+0] - dot * normals_3D[3*a+0];
        final_tangents_4D[4*a+1] = tangents[3*a+1] - dot * normals_3D[3*a+1];
        final_tangents_4D[4*a+2] = tangents[3*a+2] - dot * normals_3D[3*a+2];
        // znormalizuj tangentu
        float length = sqrt(final_tangents_4D[4*a+0]*final_tangents_4D[4*a+0]+
                            final_tangents_4D[4*a+1]*final_tangents_4D[4*a+1]+
                            final_tangents_4D[4*a+2]*final_tangents_4D[4*a+2]);
        if (length > 0)
        {
            final_tangents_4D[4*a+0] = final_tangents_4D[4*a+0] / length;
            final_tangents_4D[4*a+1] = final_tangents_4D[4*a+1] / length;
            final_tangents_4D[4*a+2] = final_tangents_4D[4*a+2] / length;
        }
        
        // do stvrtej suradnice tangenty uloz informaciu, ci lokalny tangencialny priestor je pravo alebo lavotocivy
        float cross_x = normals_3D[3*a+1]*final_tangents_4D[4*a+2] - normals_3D[3*a+2]*final_tangents_4D[4*a+1];
        float cross_y = normals_3D[3*a+2]*final_tangents_4D[4*a+0] - normals_3D[3*a+0]*final_tangents_4D[4*a+2];
        float cross_z = normals_3D[3*a+0]*final_tangents_4D[4*a+1] - normals_3D[3*a+1]*final_tangents_4D[4*a+0];
        dot = cross_x * bitangents[3*a+0] + cross_y * bitangents[3*a+1] + cross_z * bitangents[3*a+2];
        final_tangents_4D[4*a+3] = (dot < 0.0f) ? -1.0f : 1.0f;
    }
    
    delete[] tangents;
}