
// Standardne pre MS win aplikacie
#include <windows.h>

// GLEW kniznica. pre spravu novsich rozsireni
#include "glew.h"
// GL Utility Toolkit (GLUT)
#include "glut.h"

#include <stdio.h>
#include <math.h>

#include "shaders.h"
#include "textures.h"
#include "geometry.h"
#include "fonts.h"
#include "meshes.h"
#include "pose.h"
#include "fbo.h"

// Globalne premenne
bool g_is_anim = true;
bool g_is_texture = true;
float g_current_time = 0.0f;

// Globalne premenne pre prezeranie sceny
GLdouble g_vzd, g_old_vzd, g_fi, g_xi, g_old_fi, g_old_xi;
GLint g_left_mouse_down_x, g_left_mouse_down_y;
GLint g_right_mouse_down_y;
GLboolean g_right_down = false, g_left_down = false;
GLuint g_window_height, g_window_width;

// id textur
GLuint g_tex_sun, g_tex_lava, g_tex_terrain_diffuse, g_tex_skybox, g_tex_water_normal_map;

// id shaderov
GLuint g_shaders_phong, g_shaders_skybox, g_shaders_water, g_shaders_envmap;

// modely a svetla
std::vector<ObjectPose> g_light_animation_keys;
std::vector<float> g_light_animation_knots;
std::vector<ObjectPose> g_duck_animation_keys;
std::vector<float> g_duck_animation_knots;
Scene* g_duck_model = NULL;
Scene* g_car_model = NULL;
Scene* g_humvee_model = NULL;

// framebuffer objekty
FBO g_reflection_pass_fbo;
FBO g_refraction_pass_fbo;

// vlastnosti vody
std::vector<aiVector3D> g_water_plane_points;


void InitAnimations()
{
    g_duck_animation_knots.push_back(0);
    g_duck_animation_keys.push_back(CreatePose(4, 0, 0, 90, 0, 90, 0.01f, 0.01f, 0.01f));
    g_duck_animation_knots.push_back(3);
    g_duck_animation_keys.push_back(CreatePose(4, 4, 0, 90, 0, 135, 0.01f, 0.01f, 0.01f));
    g_duck_animation_knots.push_back(6);
    g_duck_animation_keys.push_back(CreatePose(-4, 4, 0, 90, 0, 220, 0.01f, 0.01f, 0.01f));
    g_duck_animation_knots.push_back(9);
    g_duck_animation_keys.push_back(CreatePose(-4, -4, 0, 90, 0, 310, 0.01f, 0.01f, 0.01f));
    g_duck_animation_knots.push_back(12);
    g_duck_animation_keys.push_back(CreatePose(4, -4, 0, 90, 0, 40, 0.01f, 0.01f, 0.01f));
    g_duck_animation_knots.push_back(15);
    g_duck_animation_keys.push_back(CreatePose(4, 0, 0, 90, 0, 90, 0.01f, 0.01f, 0.01f));

    float dist = 1;
    g_light_animation_knots.push_back(0);
    g_light_animation_keys.push_back(CreatePose(7 * dist, 0 * dist, 3 * dist, 0, 0, 0, 1, 1, 1));
    g_light_animation_knots.push_back(3);
    g_light_animation_keys.push_back(CreatePose(7 * dist, -6 * dist, 4 * dist, 0, 0, 0, 1, 1, 1));
    g_light_animation_knots.push_back(6);
    g_light_animation_keys.push_back(CreatePose(0 * dist, -7 * dist, 5 * dist, 0, 0, 0, 1, 1, 1));
    g_light_animation_knots.push_back(9);
    g_light_animation_keys.push_back(CreatePose(-7 * dist, -6 * dist, 6 * dist, 0, 0, 0, 1, 1, 1));
    g_light_animation_knots.push_back(12);
    g_light_animation_keys.push_back(CreatePose(-7 * dist, 3 * dist, 4 * dist, 0, 0, 0, 1, 1, 1));
    g_light_animation_knots.push_back(15);
    g_light_animation_keys.push_back(CreatePose(7 * dist, 0 * dist, 3 * dist, 0, 0, 0, 1, 1, 1));
}

void initShaders()
{
    // nacitanie, kompilacia a zlinkovanie shaderov
    g_shaders_phong = LoadProgram("shaders/perpixel_phong.vert", "shaders/perpixel_phong.frag");

    g_shaders_skybox = LoadProgram("shaders/skybox.vert", "shaders/skybox.frag");

    g_shaders_water = LoadProgram("shaders/water.vert", "shaders/water.frag");

    g_shaders_envmap = LoadProgram("shaders/perpixel_envmap.vert", "shaders/perpixel_envmap.frag");
}

// inicializacia OpenGL textur
void initTextures()
{
    // ************* textury pre teren *************
    g_tex_terrain_diffuse = LoadTexture("textures/terrain_diffuse.png");

    // ************* textura pre slnko *************
    g_tex_sun = LoadTexture("textures/sun.jpg");

    // ************* textura pre skybox *************
    g_tex_skybox = LoadCubeTexture("skybox/snow_positive_x.jpg", "skybox/snow_negative_x.jpg",
                                   "skybox/snow_positive_y.jpg", "skybox/snow_negative_y.jpg",
                                   "skybox/snow_positive_z.jpg", "skybox/snow_negative_z.jpg");

    // ************* textura pre normaly vody *************
    g_tex_water_normal_map = LoadTexture("textures/water_normal_map.jpg");

    g_tex_lava = LoadTexture("textures/lava.png");
}

// Open GL Inicializacia
bool initGL(void)
{
    // vypis info o systeme OpenGL
    printf("OpenGL version: %s\n", glGetString(GL_VERSION));
    printf("OpenGL vendor: %s\n", glGetString(GL_VENDOR));
    printf("OpenGL renderer: %s\n", glGetString(GL_RENDERER));

    GLint maxColorAttachments;
    glGetIntegerv(GL_MAX_COLOR_ATTACHMENTS_EXT, &maxColorAttachments);
    printf("Max FBO color attachments: %i\n", maxColorAttachments);

    GLfloat maximumAnisotropy;
    glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &maximumAnisotropy);
    printf("Max Anisotrphy filter samples: %i\n", int(maximumAnisotropy));

    GLfloat maxTexturesInVertexShader;
    glGetFloatv(GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS, &maxTexturesInVertexShader);
    printf("Max Vertex Shader Textures: %i\n", int(maxTexturesInVertexShader));

    // nastav cierne pozadie
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    
    // nastav z-buffer testovanie
    glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

    // nastav stencil buffer
    glClearStencil(0);

    // nastav zmiesavaciu funkciu
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // nastav vhodne mapovanie textur
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

    // initializuj kameru
    g_fi = 0.7f; g_xi = 0.7f; g_vzd = 10.0f;

    // pozicia vody
    g_water_plane_points.push_back(aiVector3D(-9, -9, 0));
    g_water_plane_points.push_back(aiVector3D(9, -9, 0));
    g_water_plane_points.push_back(aiVector3D(9, 9, 0));
    g_water_plane_points.push_back(aiVector3D(-9, 9, 0));


	return true;
}

// vyrenderuj celu scenu, modelview je v tomto momente uz nastavena a obsahuje view maticu kamery
void renderWorldScene(bool normal_pass)
{
    // vykresli skybox
    if (normal_pass)
    {
        // OBJEKTY POUZIVAJUCE SKYBOX SHADER
        // skybox je dany vo svetovych suradniciach, akurat jeho stred je stale v bode, kde je kamera
        glUseProgram(g_shaders_skybox);
        SetShaderUniform1i(g_shaders_skybox, "cube_map", 0);
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_CUBE_MAP, g_tex_skybox);
        glPushMatrix();
        glScalef(100, 100, 100);
        renderCubeVBO(g_shaders_skybox);
        glPopMatrix();
        glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
    }

    // OBJEKTY POUZIVAJUCE PHONG SHADER
    // nastav shadre podla toho, v ktorom passe sme
    GLuint current_shader = g_shaders_phong;
    glUseProgram(current_shader);

    // vykresli slnko
    SetShaderUniform1i(current_shader, "lighting_enabled", 0);
    SetShaderUniform1i(current_shader, "texturing_enabled", 1);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, g_tex_sun);
    glPushMatrix();
    ObjectPose light_pose = IntepolatePoseCubic(g_light_animation_knots, g_light_animation_keys, g_current_time, false);
    MultiplyMatrixOpenGL(GetMatrix(light_pose));
    GLfloat sun_position[] = { 0.0f, 0.0f, 0.0f, 1.0f };
    glLightfv(GL_LIGHT0, GL_POSITION, sun_position);
    glScalef(0.2f, 0.2f, 0.2f);
    glColor3f(1.0f, 1.0f, 1.0f);
    renderSphereVBO(current_shader);
    glPopMatrix();

    // vykresli humvee
    SetShaderUniform1i(current_shader, "lighting_enabled", 1);
    glPushMatrix();
    ObjectPose humvee_pose = CreatePose(2, 0, 0.0, 90, 0, 0, 0.008f, 0.008f, 0.008f);
    MultiplyMatrixOpenGL(GetMatrix(humvee_pose)); 
    RenderScene(g_humvee_model, current_shader);
    glPopMatrix();

    // vykresli kacku
    glPushMatrix();
    ObjectPose duck_pose = IntepolatePoseCubic(g_duck_animation_knots, g_duck_animation_keys, g_current_time, false);
    MultiplyMatrixOpenGL(GetMatrix(duck_pose)); 
    RenderScene(g_duck_model, current_shader);
    glPopMatrix();

    // vykresli auto
    glPushMatrix();
    ObjectPose car_pose = CreatePose(-2, 0, -1.5, 90, 0, 0, 0.35f, 0.35f, 0.35f);
    MultiplyMatrixOpenGL(GetMatrix(car_pose)); 
    RenderScene(g_car_model, current_shader);
    glPopMatrix();

    // vykresli 3D text
    SetShaderUniform1i(current_shader, "texturing_enabled", 0);
    GLfloat text_diffuse_material[] = {1.0f, 0.0f, 0.0f, 1.0f};
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, text_diffuse_material);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, text_diffuse_material);
    glColor3f(1.0f, 0.0f, 0.0f);
    render3DTextWGL(6, 0, 2, 1, "Project GRC - Reflections, Refractions ...");
    glDisable(GL_LIGHTING);

    // // OBJEKTY POUZIVAJUCE ENVIRONMENT MAPPING SHADER
    glUseProgram(g_shaders_envmap);
    // vykresli torus s environment mapou
    // najprv nastav uniformy pre envmap shader
    glPushMatrix();
    GLfloat camera_view_matrix[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, camera_view_matrix);
    SetShaderUniformMatrix4f(g_shaders_envmap, "camera_view_matrix", camera_view_matrix);
    SetShaderUniform1i(g_shaders_envmap, "diffuse_map", 0);
    SetShaderUniform1i(g_shaders_envmap, "cube_map", 1);
    SetShaderUniform1i(g_shaders_envmap, "texturing_enabled", 1);
    GLfloat camera_pos[4];
    camera_pos[0] = g_vzd * cos(g_fi) * cos(g_xi);
	camera_pos[1] = g_vzd * sin(g_fi) * cos(g_xi);
	camera_pos[2] = g_vzd * sin(g_xi);
    camera_pos[3] = 1;
    SetShaderUniform4f(g_shaders_envmap, "camera_world_pos", camera_pos);
    // vynuluj modelview, od teraz je v aktualnej modelview matici len model cast a view cast sme si poslali do sahdera ako uniform
    glLoadIdentity();
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, g_tex_lava);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_CUBE_MAP, g_tex_skybox);
    ObjectPose torus_pose = CreatePose(-2, 0, 1.1, 0, 0, 0, 1.05f, 1.05f, 1.05f);
    MultiplyMatrixOpenGL(GetMatrix(torus_pose)); 
    renderTorusVBO(g_shaders_envmap);
    glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
    glActiveTexture(GL_TEXTURE0);
    glPopMatrix();
}

// Procedura obsluhujuca vykreslovanie
void render(void)   
{
    GLfloat camera_pos[4];
    camera_pos[0] = g_vzd * cos(g_fi) * cos(g_xi);
	camera_pos[1] = g_vzd * sin(g_fi) * cos(g_xi);
	camera_pos[2] = g_vzd * sin(g_xi);
    camera_pos[3] = 1;

    // 1.PASS - RENDEROVANIE SCENY SPOD ROVINY VODY
    // Nastav FBO pre renderovanie depth mapy do textury
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, g_reflection_pass_fbo.id);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    // Nastav viewport na velkost okna
    glViewport(0, 0, g_window_width, g_window_height);
    // Nastav projekcnu maticu, je rovnaka ako pre klasicku kameru
    glMatrixMode(GL_PROJECTION);                        
	glLoadIdentity();
    gluPerspective(60, g_window_width / float(g_window_height), 0.1, 300.0);
    // Nastav view maticu pre vyrenderovanie reflekcnej matice z druhej strany vodnej hladiny
    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
    ObjectPose reflection_pose = CreateMirrorLookAtPose(camera_pos[0], camera_pos[1], camera_pos[2], 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, g_water_plane_points);
    MultiplyMatrixOpenGL(GetMatrix(reflection_pose).Inverse());
    // Nastav orezavaciu rovinu, aby sme renderovali len objekty nad hladinou
    GLdouble plane_equation[4];
    plane_equation[0] = plane_equation[1] = plane_equation[2] = 0.0;
    for (int i = 0; i < (int)g_water_plane_points.size(); i++) 
    {
        int j = (i + 1) % (int)g_water_plane_points.size();
        plane_equation[0] += (g_water_plane_points[i].y - g_water_plane_points[j].y) * (g_water_plane_points[i].z + g_water_plane_points[j].z);
        plane_equation[1] += (g_water_plane_points[i].z - g_water_plane_points[j].z) * (g_water_plane_points[i].x + g_water_plane_points[j].x);
        plane_equation[2] += (g_water_plane_points[i].x - g_water_plane_points[j].x) * (g_water_plane_points[i].y + g_water_plane_points[j].y);
    }
    plane_equation[3] = -(plane_equation[0]*g_water_plane_points[0].x + plane_equation[1]*g_water_plane_points[0].y + plane_equation[2]*g_water_plane_points[0].z);
    glClipPlane(GL_CLIP_PLANE0, plane_equation);
    glEnable(GL_CLIP_PLANE0);
    // vykresli scenu spod vody
    renderWorldScene(false);

    // 2.PASS - RENDEROVANIE Z POHLADU KAMERY DO REFRAKCNEJ TEXTURY
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, g_refraction_pass_fbo.id);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	
    // Nastav viewport kamery
    glViewport(0, 0, g_window_width, g_window_height);
    // Nastav projekcnu maticu kamery
    glMatrixMode(GL_PROJECTION);                        
	glLoadIdentity();
    gluPerspective(60, g_window_width / float(g_window_height), 0.1, 300.0);
    // Nastav view maticu kamery
    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
    ObjectPose refraction_pose = CreateLookAtPose(camera_pos[0], camera_pos[1], camera_pos[2], 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f);
    MultiplyMatrixOpenGL(GetMatrix(refraction_pose).Inverse());
    // nastav opacne orezavanie ako v pripade reflekcie
    plane_equation[0] = -plane_equation[0];
    plane_equation[1] = -plane_equation[1];
    plane_equation[2] = -plane_equation[2];
    plane_equation[3] = -plane_equation[3];
    glClipPlane(GL_CLIP_PLANE0, plane_equation);
    glEnable(GL_CLIP_PLANE0);
    // vykresli scenu
    renderWorldScene(false);
    glDisable(GL_CLIP_PLANE0);

    // 3.PASS - NORMALNE RENDEROVANIE Z POHLADU KAMERY DO OKNA
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    // Nastav viewport na velkost okna
    glViewport(0, 0, g_window_width, g_window_height);
    // Nastav projekcnu maticu klasickej kamery
    glMatrixMode(GL_PROJECTION);                        
	glLoadIdentity();
    gluPerspective(60, g_window_width / float(g_window_height), 0.1, 300.0);
    // Nastav view maticu pre vyrenderovanie reflekcnej matice z druhej strany vodnej hladiny
    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
    ObjectPose camera_pose = CreateLookAtPose(camera_pos[0], camera_pos[1], camera_pos[2], 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f);
    MultiplyMatrixOpenGL(GetMatrix(camera_pose).Inverse());
    GLfloat camera_view_matrix[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, camera_view_matrix);
    renderWorldScene(true);

    // VYKRESLENIE VODNEJ HLADINY
    // Nastav water shader
    glUseProgram(g_shaders_water);
    // Nastav uniformy
    SetShaderUniformMatrix4f(g_shaders_water, "camera_view_matrix", camera_view_matrix);
    SetShaderUniform4f(g_shaders_water, "camera_world_pos", camera_pos);
    ObjectPose light_pose = IntepolatePoseCubic(g_light_animation_knots, g_light_animation_keys, g_current_time, false);
    GLfloat light_pos[4];
    light_pos[0] = light_pose.pos_x;
    light_pos[1] = light_pose.pos_y;
    light_pos[2] = light_pose.pos_z;
    light_pos[3] = 1.0f;
    SetShaderUniform4f(g_shaders_water, "light_world_pos", light_pos);
    GLfloat tangent_world[4];
    tangent_world[0] = 1.0f;
    tangent_world[1] = tangent_world[2] = tangent_world[3] = 0.0f;
    SetShaderUniform4f(g_shaders_water, "tangent_world", tangent_world);
    GLfloat normal_world[4];
    normal_world[2] = 1.0f;
    normal_world[0] = normal_world[1] = normal_world[3] = 0.0f;
    SetShaderUniform4f(g_shaders_water, "normal_world", normal_world);
    SetShaderUniform1f(g_shaders_water, "window_width", g_window_width);
    SetShaderUniform1f(g_shaders_water, "window_height", g_window_height);
    SetShaderUniform1f(g_shaders_water, "rippling_strength", 0.05f);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, g_reflection_pass_fbo.color_textures[0]);
    SetShaderUniform1i(g_shaders_water, "reflection_map", 0);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_2D, g_refraction_pass_fbo.color_textures[0]);
    SetShaderUniform1i(g_shaders_water, "refraction_map", 1);
    glActiveTexture(GL_TEXTURE2);
    glBindTexture(GL_TEXTURE_2D, g_tex_water_normal_map);
    SetShaderUniform1i(g_shaders_water, "normal_map", 2);
    glActiveTexture(GL_TEXTURE3);
    glBindTexture(GL_TEXTURE_CUBE_MAP, g_tex_skybox);
    SetShaderUniform1i(g_shaders_water, "cube_map", 3);
    GLfloat light_diffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    // vykresli vodnu hladinu
    glLoadIdentity();
    glBegin(GL_POLYGON);
    for (int i = 0; i < (int)g_water_plane_points.size(); i++)
    {
        glTexCoord2f(g_water_plane_points[i].x / 3.0f, g_water_plane_points[i].y / 3.0f);
        glVertex3f(g_water_plane_points[i].x, g_water_plane_points[i].y, g_water_plane_points[i].z);
    }
    glEnd();
    glActiveTexture(GL_TEXTURE3);
    glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
    glActiveTexture(GL_TEXTURE2);
    glBindTexture(GL_TEXTURE_2D, 0);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_2D, 0);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, 0);

    // vykresli mapy do okna
    if (g_is_texture)
    {
        glUseProgram(NULL);
        glDisable(GL_LIGHTING);
        glColor3f(1.0f, 1.0f, 1.0f);
        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glLoadIdentity();
        glOrtho(0, g_window_width, 0, g_window_height, -1, 1);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        glActiveTexture(GL_TEXTURE0);
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, g_reflection_pass_fbo.color_textures[0]);
        glBegin(GL_QUADS);
        glTexCoord2f(0, 0);glVertex3f(0,0,0);
        glTexCoord2f(1, 0);glVertex3f(0.3f*g_window_width,0,0);
        glTexCoord2f(1, 1);glVertex3f(0.3f*g_window_width,0.3f*g_window_height,0);
        glTexCoord2f(0, 1);glVertex3f(0,0.3f*g_window_height,0);
        glEnd();
        glBindTexture(GL_TEXTURE_2D, g_refraction_pass_fbo.color_textures[0]);
        glBegin(GL_QUADS);
        glTexCoord2f(0, 0);glVertex3f(0.3f*g_window_width,0,0);
        glTexCoord2f(1, 0);glVertex3f(0.6f*g_window_width,0,0);
        glTexCoord2f(1, 1);glVertex3f(0.6f*g_window_width,0.3f*g_window_height,0);
        glTexCoord2f(0, 1);glVertex3f(0.3f*g_window_width,0.3f*g_window_height,0);
        glEnd();
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
    }

    // vykresli 2D text, bez shaderov
    glUseProgram(NULL);
    glActiveTexture(GL_TEXTURE0);
    glDisable(GL_LIGHTING);
    glColor3f(1.0f, 1.0f, 0.0f);
    render2DTextFT(10, g_window_height - 20, "Projekt GRC ...");
    render2DTextFT(10, g_window_height - 45, "F1 - zapni/vypni animaciu");
    render2DTextFT(10, g_window_height - 70, "F2 - zapni/vypni mapy");

    // Zamen buffre a zobraz vykreslene objekty
    glutSwapBuffers();
}

// Procedura obsluhujuca zmenu rozmerov okna
void reshape(int w, int h)
{
    // nastav FBO ktore maju velkost okna
    g_reflection_pass_fbo.Clear();
    g_reflection_pass_fbo = CreateFBO(w, h, 1, false);
    g_refraction_pass_fbo.Clear();
    g_refraction_pass_fbo = CreateFBO(w, h, 1, false);

    g_window_width = w;
    g_window_height = h;
	if (g_window_height == 0) 
        g_window_height = 1;
}

// Procedura obsluhujuca stlacenie klavesy
void keyboard(unsigned char key, int x, int y)
{
	switch (key) 
    {
        // Ak Esc bolo stlacene
		case 27:
            // vymazalokovane OpenGL data
            DeleteAllVBO();
            if (g_duck_model)
                delete g_duck_model;
            if (g_car_model)
                delete g_car_model;
            if (g_humvee_model)
                delete g_humvee_model;
            g_reflection_pass_fbo.Clear();
            g_refraction_pass_fbo.Clear();
            // Ukoncenie programu
			exit(0);
		    break;
		default:        
		    break;
	}
}

// Procedura obsluhujuca stlacenie specialnej klavesy
void special_keys(int a_keys, int x, int y)
{
	switch (a_keys) 
    {
        // Ak F1 bolo stlacene
		case GLUT_KEY_F1:
            g_is_anim = !g_is_anim;
            glutPostRedisplay();
		    break;
        // Ak F2 bolo stlacene
        case GLUT_KEY_F2:
            g_is_texture = !g_is_texture;
            glutPostRedisplay();
		    break;
		default:
		    break;
	}
}

// Procedura obsluhujuca stlacenie tlacitka mysi
void mouse_klik(int button, int state, int x, int y )
{
	switch (button) 
    {
		case GLUT_LEFT_BUTTON:
            if (state == GLUT_DOWN)
            {
                g_left_down = true;
			    g_left_mouse_down_x = x;
			    g_left_mouse_down_y = y;
			    g_old_fi = g_fi;
			    g_old_xi = g_xi;
                break;
            }
            else if (state == GLUT_UP)
            {
                g_left_down = false;
            }
		    break;
        case GLUT_RIGHT_BUTTON:
            if (state == GLUT_DOWN)
            {
                g_right_down = true;
			    g_right_mouse_down_y = y;
			    g_old_vzd = g_vzd;
                break;
            }
            else if (state == GLUT_UP)
            {
                g_right_down = false;
            }
		    break;
		default:
		    break;
	}

    glutPostRedisplay();
}

// Procedura obsluhujuca pohyb mysi
void mouse_move(int x, int y)
{
    if (g_left_down == GL_TRUE)
    {
        g_fi = g_old_fi + (g_left_mouse_down_x - x)/200.0f;
        g_xi = g_old_xi + (y - g_left_mouse_down_y)/200.0f;
        glutPostRedisplay();
    }

    if (g_right_down == GL_TRUE)
    {
        g_vzd = g_old_vzd + (g_right_mouse_down_y - y)/10.0f;
        if (g_vzd > 80) g_vzd = 80;
        glutPostRedisplay();
    } 

}

void timer(int id)
{
    if (g_is_anim)
    {
        g_current_time += 0.050f;
    }
    else
    {
        //g_current_time = 0.0f;
    }

    // jednoduche opakovanie animacie po 17 sekundach
    if (g_current_time > 17)
        g_current_time = 0.0f;

    glutPostRedisplay();
    glutTimerFunc(50, timer, 0);
}

// Zakladna funkcia spustena na zaciatku
int main(int argc, char** argv)
{
    // GLUT Inicializacia
	glutInit(&argc, argv);                              
    // Display Mod (Rgb a Double Buffering)
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);
    //glutInitDisplayString("stencil>=8 rgb>=8 double depth>=24");
    // Nastavime pociatocnu poziciu a velkost okna
    glutInitWindowPosition(50, 50);
    glutInitWindowSize(800, 600);                   
    // Vyrobime okno s danym menom
    glutCreateWindow("--== GRC projekt ==--" );  

    // Zinicializujeme GLEW
    glewInit();

    // Nastavime zakladne vlastnosti OpenGL
	initGL();
    initSphereVBO(32, 32);
    initTorusVBO(0.6f, 2.0f, 32, 32);
    initTextures();
    initShaders();
    init2DFontFT("fonts/optien.ttf");
    InitAnimations();

    // nacitaj modely
    g_duck_model = LoadScene("./models/Duck/duck.dae");
    g_car_model = LoadScene("./models/F40/F40.dae");
    g_humvee_model = LoadScene("./models/Humvee/HMMWV.3ds");

    // Zaregistrujeme callback funkcie
	glutDisplayFunc(render);                            
	glutReshapeFunc(reshape);                        
	glutKeyboardFunc(keyboard);                      
	glutSpecialFunc(special_keys);
    glutMouseFunc(mouse_klik);
    glutMotionFunc(mouse_move);
    glutTimerFunc(0, timer, 0);

    // Spustenie hlavneho okruhu zachytavania sprav
	glutMainLoop();

	return 0;
}
