varying vec4 V_world;
varying vec4 L_world;

uniform sampler2D reflection_map;
uniform sampler2D refraction_map;
uniform sampler2D normal_map;
uniform samplerCube cube_map;
uniform float window_width, window_height;
uniform float rippling_strength;
uniform vec4 tangent_world;
uniform vec4 normal_world;

void main(void)
{
   // get new normal from normal map in world coordinates
   vec3 normal = vec3(2.0 * texture2D(normal_map, gl_TexCoord[0].st) - 1.0);
   vec3 bitangent_world = cross(normal_world.xyz, tangent_world.xyz);
   mat3 TBN_world = mat3(tangent_world.xyz, bitangent_world, normal_world.xyz);
   vec4 new_normal_world = normalize(vec4(TBN_world * normal, 0.0));
   
   // compute texture coordinates for reflection and refraction textures 
   vec2 reflection_tex_coord = vec2(gl_FragCoord.x / window_width, gl_FragCoord.y / window_height);
   vec2 refraction_tex_coord = reflection_tex_coord + normal.xz * rippling_strength / gl_FragCoord.z;
   reflection_tex_coord = reflection_tex_coord + normal.xz * rippling_strength / gl_FragCoord.z;
   reflection_tex_coord.x = 1 - reflection_tex_coord.x;
   
   // get local reflection and refraction colors
   vec4 reflection_color = texture2D(reflection_map, reflection_tex_coord);
   vec4 refraction_color = texture2D(refraction_map, refraction_tex_coord);
   
   // get reflection vector into environment map
   vec4 V = normalize(V_world);
   vec4 R_world = reflect(-V, new_normal_world);
   vec4 env_reflection = textureCube(cube_map, R_world.xzy);
   
   // get sun reflection
   float VdotL = dot(V, reflect(-normalize(L_world), new_normal_world));
   vec4 sun_reflection = pow(clamp(VdotL, 0.0, 1.0), 32.0) * gl_LightSource[0].diffuse;
   
   // get fresnel
   float fresnel = 1.0 - pow(clamp(dot(new_normal_world, V), 0.0, 1.0), 5.0);

   gl_FragColor = (1.0 - fresnel) * refraction_color;
   gl_FragColor += fresnel * (0.8 * sun_reflection);
   gl_FragColor += fresnel * (0.6 * env_reflection);
   gl_FragColor += fresnel * (0.3 * reflection_color);
   gl_FragColor.a = 1.0;
   //gl_FragColor = sun_reflection + env_reflection;
}
